#region Copyright & License Information
/*
 * Copyright 2007-2011 The OpenRA Developers (see AUTHORS)
 * This file is part of OpenRA, which is free software. It is made
 * available to you under the terms of the GNU General Public License
 * as published by the Free Software Foundation. For more information,
 * see COPYING.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using OpenRA.FileFormats;
using OpenRA.Mods.RA.Buildings;
using OpenRA.Mods.RA.Move;
using OpenRA.Traits;
using XRandom = OpenRA.Thirdparty.Random;
using System.Collections;
using OpenRA.Mods.RA.Air;

//TODO:
// effectively clear the area around the production buildings' spawn points.
// don't spam the build unit button, only queue one unit then wait for the backoff period.
//    just make the build unit action only occur once every second.

// later:
// don't build units randomly, have a method to it.
// explore spawn points methodically
// once you find a player, attack the player instead of spawn points.

namespace OpenRA.Mods.RA.AI
{
    class HackyAIInfo : IBotInfo, ITraitInfo
    {
        public readonly string Name = "Unnamed Bot";
        public readonly int SquadSize = 8;
        public readonly int AssignRolesInterval = 20;
        public readonly string RallypointTestBuilding = "fact";		// temporary hack to maintain previous rallypoint behavior.
        public readonly string[] UnitQueues = { "Vehicle", "Infantry", "Plane" };
        public readonly bool ShouldRepairBuildings = true;

        string IBotInfo.Name { get { return this.Name; } }

        [FieldLoader.LoadUsing("LoadUnits")]
        public readonly Dictionary<string, float> UnitsToBuild = null;

        [FieldLoader.LoadUsing("LoadBuildings")]
        public readonly Dictionary<string, float> BuildingFractions = null;

        static object LoadActorList(MiniYaml y, string field)
        {
            return y.NodesDict[field].Nodes.ToDictionary(
                t => t.Key,
                t => FieldLoader.GetValue<float>(field, t.Value.Value));
        }

        static object LoadUnits(MiniYaml y) { return LoadActorList(y, "UnitsToBuild"); }
        static object LoadBuildings(MiniYaml y) { return LoadActorList(y, "BuildingFractions"); }

        public object Create(ActorInitializer init) { return new HackyAI(this); }
    }

    /* a pile of hacks, which control a local player on the host. */

    class Enemy { public int Aggro; }

    class HackyAI : ITick, IBot, INotifyDamage
    {
        bool enabled;
        public int ticks;
        public Player p;
        PowerManager playerPower;
        PlayerResources playerResource;
        readonly BuildingInfo rallypointTestBuilding;		// temporary hack
        readonly HackyAIInfo Info;

        Cache<Player, Enemy> aggro = new Cache<Player, Enemy>(_ => new Enemy());
        CPos baseCenter;
        XRandom random = new XRandom(); //we do not use the synced random number generator.
        BaseBuilder[] builders;

        // AI improvement
        const int Max_DOME = 1;
        const int Max_PROC = 2;
        const int Max_STEK = 1;
        const int Max_ATEK = 1;
        const int Max_SILO = 1;
        const int Max_FIX = 1;

        const int Max_BARR = 2;
        const int Max_TENT = 2;
        const int Max_WEAP = 2;

        const int OverrunRation = 2;

        const int MaxBaseDistance = 40;
        public const int feedbackTime = 30;		// ticks; = a bit over 1s. must be >= netlag.

        public World world { get { return p.PlayerActor.World; } }
        IBotInfo IBot.Info { get { return this.Info; } }

        public HackyAI(HackyAIInfo Info)
        {
            this.Info = Info;
            // temporary hack.
            this.rallypointTestBuilding = Rules.Info[Info.RallypointTestBuilding].Traits.Get<BuildingInfo>();
        }

        public static void BotDebug(string s, params object[] args)
        {
            //if (Game.Settings.Debug.BotDebug)
                Game.Debug(s, args);
        }

        /* called by the host's player creation code */
        public void Activate(Player p)
        {
            this.p = p;
            enabled = true;
            playerPower = p.PlayerActor.Trait<PowerManager>();
            playerResource = p.PlayerActor.Trait<PlayerResources>();
            builders = new BaseBuilder[] {
				new BaseBuilder( this, "Building", q => ChooseBuildingToBuild(q, true) ),
				new BaseBuilder( this, "Defense", q => ChooseBuildingToBuild(q, false) ) };
            /* TEST TAUNT */
            Order.Chat(false, "Good luck.");
        }

        int GetPowerProvidedBy(ActorInfo building)
        {
            var bi = building.Traits.GetOrDefault<BuildingInfo>();
            if (bi == null) return 0;
            return bi.Power;
        }

        ActorInfo ChooseRandomUnitToBuild(ProductionQueue queue)
        {
            var buildableThings = queue.BuildableItems();
            if (!buildableThings.Any()) return null;
            return buildableThings.ElementAtOrDefault(random.Next(buildableThings.Count()));
        }

        bool HasAdequatePower()
        {
            /* note: CNC `fact` provides a small amount of power. don't get jammed because of that. */
            return playerPower.PowerProvided > 50 &&
                playerPower.PowerProvided > playerPower.PowerDrained * 1.2;
        }

        int countBuilding(string frac, Player owner)
        {
            return owner.World.ActorsWithTrait<Building>().Where(a => a.Actor.Owner == owner && a.Actor.Info.Name == frac).Count();
        }

        public bool HasAdequateNumber(string frac, Player owner)
        {
            int count = countBuilding(frac, owner);
            if (frac == "dome" && count >= Max_DOME)
                return false;
            if (frac == "proc" && count >= Max_PROC)
                return false;
            if (frac == "stek" && count >= Max_STEK)
                return false;
            if (frac == "atek" && count >= Max_ATEK)
                return false;
            if (frac == "barr" && count >= Max_BARR)
                return false;
            if (frac == "tent" && count >= Max_TENT)
                return false;
            if (frac == "weap" && count >= Max_WEAP)
                return false;
            //if (frac == "silo" && count >= Max_SILO * countBuilding("proc", owner))
            //    return false;
            if (frac == "fix" && count >= Max_FIX * countBuilding("weap", owner))
                return false;
            if (frac == "hpad" && count > 0)
                if(count != p.World.ActorsWithTrait<Helicopter>().Where(a => a.Actor.Owner == owner).Count())
                    return false;

            return true;
        }

        ActorInfo ChooseBuildingToBuild(ProductionQueue queue, bool buildPower)
        {
            var buildableThings = queue.BuildableItems();

            if (!HasAdequatePower())	/* try to maintain 20% excess power */
            {
                if (!buildPower) return null;

                /* find the best thing we can build which produces power */
                return buildableThings.Where(a => GetPowerProvidedBy(a) > 0)
                    .OrderByDescending(a => GetPowerProvidedBy(a)).FirstOrDefault();
            }

            var myBuildings = p.World
                .ActorsWithTrait<Building>()
                .Where(a => a.Actor.Owner == p)
                .Select(a => a.Actor.Info.Name).ToArray();

            foreach (var frac in Info.BuildingFractions)
            {
                if (buildableThings.Any(b => b.Name == frac.Key))
                    if (myBuildings.Count(a => a == frac.Key) < frac.Value * myBuildings.Length && playerPower.ExcessPower >= Rules.Info[frac.Key].Traits.Get<BuildingInfo>().Power) //if (random.Next(100) <= frac.Value && playerPower.ExcessPower >= Rules.Info[frac.Key].Traits.Get<BuildingInfo>().Power)
                        if (HasAdequateNumber(frac.Key, p)) /* C'mon... */
                            if(!playerResource.AlertSilo)
                                return Rules.Info[frac.Key];
                            else
                                return Rules.Info["silo"]; /* C'mon... */
            }
            return null;
        }

        bool NoBuildingsUnder(IEnumerable<CPos> cells)
        {
            var bi = world.WorldActor.Trait<BuildingInfluence>();
            return cells.All(c => bi.GetBuildingAt(c) == null);
        }

        public CPos? ChooseBuildLocation(string actorType)
        {
            var bi = Rules.Info[actorType].Traits.Get<BuildingInfo>();

            for (var k = 0; k < MaxBaseDistance; k++)
                foreach (var t in world.FindTilesInCircle(baseCenter, k))
                    if (world.CanPlaceBuilding(actorType, bi, t, null))
                        if (bi.IsCloseEnoughToBase(world, p, actorType, t))
                            if (NoBuildingsUnder(Util.ExpandFootprint(
                                FootprintUtils.Tiles(actorType, bi, t), false)))
                                return t;

            return null;		// i don't know where to put it.
        }

        public void Tick(Actor self)
        {
            if (!enabled)
                return;

            ticks++;

            if (ticks == 1)
                DeployMcv(self);

            if (ticks % feedbackTime == 0)
                foreach (var q in Info.UnitQueues)
                    BuildRandom(q);

            AssignRolesToIdleUnits(self);
            SetRallyPointsForNewProductionBuildings(self);

            foreach (var b in builders)
                b.Tick();
        }

        //hacks etc sigh mess.
        //A bunch of hardcoded lists to keep track of which units are doing what.
        List<Actor> unitsHangingAroundTheBase = new List<Actor>();
        List<Actor> attackForce = new List<Actor>();
        List<Actor> unitsReadyToLoad = new List<Actor>();
        CPos? attackTarget;

        //Units that the ai already knows about. Any unit not on this list needs to be given a role.
        List<Actor> oldUnits = new List<Actor>();

        CPos? ChooseEnemyTarget()
        {
            // I should be able to use this to manipulate AI behavior around enemies way of playing.
            var liveEnemies = world.Players
                .Where(q => p != q && p.Stances[q] == Stance.Enemy)
                .Where(q => p.WinState == WinState.Undefined && q.WinState == WinState.Undefined);

            var leastLikedEnemies = liveEnemies
                .GroupBy(e => aggro[e].Aggro)
                .OrderByDescending(g => g.Key)
                .FirstOrDefault();

            if (leastLikedEnemies == null)
                return null;

            var enemy = leastLikedEnemies.Random(random);

            /* pick something worth attacking owned by that player */
            var targets = world.Actors
                .Where(a => a.Owner == enemy && a.HasTrait<IOccupySpace>());
            Actor target = null;

            if (targets.Any())
                target = targets.Random(random);

            if (target == null)
            {
                /* Assume that "enemy" has nothing. Cool off on attacks. */
                aggro[enemy].Aggro = aggro[enemy].Aggro / 2 - 1;
                Log.Write("debug", "Bot {0} couldn't find target for player {1}", this.p.ClientIndex, enemy.ClientIndex);

                return null;
            }

            /* bump the aggro slightly to avoid changing our mind */
            if (leastLikedEnemies.Count() > 1)
                aggro[enemy].Aggro++;

            return target.Location;
        }

        int assignRolesTicks = 0;

        void AssignRolesToIdleUnits(Actor self)
        {
            //HACK: trim these lists -- we really shouldn't be hanging onto all this state
            //when it's invalidated so easily, but that's Matthew/Alli's problem.
            oldUnits.RemoveAll(a => a.Destroyed);
            unitsHangingAroundTheBase.RemoveAll(a => a.Destroyed);
            attackForce.RemoveAll(a => a.Destroyed);
            unitsReadyToLoad.RemoveAll(a => a.Destroyed);

            if (--assignRolesTicks > 0)
                return;
            else
                assignRolesTicks = Info.AssignRolesInterval;

            var Units = self.World.ActorsWithTrait<IMove>().Where(a => a.Actor.Owner == p && !a.Actor.HasTrait<BaseBuilding>()).Select(a => a.Actor).ToArray();

            foreach (var a in Units)
            {
                var enemynearby = p.World.FindUnitsInCircle(a.CenterLocation, 3).Where(a1 => p.Stances[a1.Owner] == Stance.Enemy && !a1.Destroyed);
                if (enemynearby.Count() != 0)
                    world.IssueOrder(new Order("Attack", a, false) { TargetActor = enemynearby.First() }); // Attack ? AttackMove ?
            }

            // Find idle harvesters and give them orders:
            foreach (var a in oldUnits)
            {
                var harv = a.TraitOrDefault<Harvester>();
                if (harv == null) continue;
                if (!a.IsIdle)
                {
                    Activity act = a.GetCurrentActivity();
                    // A Wait activity is technically idle:
                    if ((act.GetType() != typeof(OpenRA.Mods.RA.Activities.Wait)) &&
                        (act.NextActivity == null || act.NextActivity.GetType() != typeof(OpenRA.Mods.RA.Activities.FindResources)))
                        continue;
                }
                if (!harv.IsEmpty) continue;

                // Tell the idle harvester to quit slacking:
                world.IssueOrder(new Order("Harvest", a, false));
            }

            var newUnits = self.World.ActorsWithTrait<IMove>()
                .Where(a => a.Actor.Owner == p && !a.Actor.HasTrait<BaseBuilding>()
                    && !oldUnits.Contains(a.Actor))
                    .Select(a => a.Actor).ToArray();

            foreach (var a in newUnits)
            {
                if (a.HasTrait<Harvester>())
                    world.IssueOrder(new Order("Harvest", a, false));
                else if (a.Info.Name.Equals("tran")) //else if (a.HasTrait<Cargo>()) /* Yet I'm focusing on TRANSPORT COPTER */
                    unitsReadyToLoad.Add(a);
                else
                    unitsHangingAroundTheBase.Add(a);

                oldUnits.Add(a);
            }

            int randomizedSquadSize = Info.SquadSize -4 + random.Next(20);
            
            if (unitsHangingAroundTheBase.Count >= randomizedSquadSize)
            {
                //if (attackForce.Count == 0)
                //{
                    attackTarget = ChooseEnemyTarget();
                    if (attackTarget == null)
                        return;

                    foreach (var a in unitsHangingAroundTheBase)
                        attackForce.Add(a);

                    unitsHangingAroundTheBase.Clear();
                //}
            }

            if (attackForce.Count > 0)
            {
                bool foundEnemy = false;

                var enemyUnits = world.Actors.Where(unit => p.Stances[unit.Owner] == Stance.Enemy && unit.HasTrait<IMove>()).ToList(); // Common
                var enemyUnitsCapturable = world.Actors.Where(unit => p.Stances[unit.Owner] == Stance.Enemy && unit.HasTrait<Capturable>()).ToList(); // Capturable
                var enemyUnitsHijack = world.Actors.Where(unit => p.Stances[unit.Owner] == Stance.Enemy && unit.HasTrait<IAcceptSpy>()).ToList(); // Hijack
                var enemyUnitsDisguise = enemyUnits.Where(unit => unit.Info.Name.Equals("e6") || unit.Info.Name.Equals("e1") || unit.Info.Name.Equals("e3"));

                if (enemyUnits.Count > 0)
                {
                    foundEnemy = true;

                    foreach (var a1 in attackForce)
                    {
                        var enemy = enemyUnits.ClosestTo(a1.CenterLocation);
                        var enemydisguise = enemyUnitsDisguise.ClosestTo(a1.CenterLocation);
                        var enemycapturable = enemyUnitsCapturable.ClosestTo(a1.CenterLocation);
                        var enemyhijack = enemyUnitsHijack.ClosestTo(a1.CenterLocation);

                        if (a1.Info.Name.Equals("e6")) // ENGINEER: CLIMB ON BOARD / MOVE TO / CAPTURE
                        {
                            foreach (var a2 in unitsReadyToLoad)
                                if (a2.TraitOrDefault<Cargo>().CanLoad(a2, a1) && a1.World.FindUnitsInCircle(a1.CenterLocation, 12).Contains(a2))
                                {
                                    world.IssueOrder(new Order("EnterTransport", a1, true) { TargetActor = a2 });
                                    break;
                                }
                            if (enemycapturable != null)
                            {
                                world.IssueOrder(new Order("Move", a1, false) { TargetLocation = enemycapturable.CenterLocation.ToCPos() });
                                world.IssueOrder(new Order("CaptureActor", a1, true) { TargetActor = enemycapturable });
                            }
                        }

                        else if (a1.Info.Name.Equals("spy")) // SPY: DISGUISE / CLIMB ON BOARD / MOVE TO / INFLITRATE
                        {
                            if(enemydisguise != null)
                                world.IssueOrder(new Order("Disguise", a1, true) { TargetActor = enemydisguise });
                            foreach (var a2 in unitsReadyToLoad)
                                if (a2.TraitOrDefault<Cargo>().CanLoad(a2, a1) && a1.World.FindUnitsInCircle(a1.CenterLocation, 12).Contains(a2))
                                {
                                    world.IssueOrder(new Order("EnterTransport", a1, true) { TargetActor = a2 });
                                    break;
                                }
                            if (enemyhijack != null)
                            {
                                world.IssueOrder(new Order("Move", a1, false) { TargetLocation = enemyhijack.CenterLocation.ToCPos() });
                                world.IssueOrder(new Order("SpyInfiltrate", a1, true) { TargetActor = enemyhijack });
                            }
                        }

                        else if (a1.Info.Name.Equals("e1") || a1.Info.Name.Equals("e3")) // SOLDIER: CLIMB ON BOARD / MOVE TO / ATTACK
                        {
                            foreach (var a2 in unitsReadyToLoad)
                                if (a2.TraitOrDefault<Cargo>().CanLoad(a2, a1) && a1.World.FindUnitsInCircle(a1.CenterLocation, 12).Contains(a2))
                                {
                                    world.IssueOrder(new Order("EnterTransport", a1, true) { TargetActor = a2 });
                                    break;
                                }
                                world.IssueOrder(new Order("Move", a1, false) { TargetLocation = enemy.CenterLocation.ToCPos() });
                                world.IssueOrder(new Order("Attack", a1, true) { TargetActor = enemy }); // Attack ? AttackMove ?
                        }

                        else if (a1.Info.Name.Equals("tran")) // TRANSPORT: MOVE TO / UNLOAD / RETURN TO BASE
                        {
                            if (!a1.TraitOrDefault<Cargo>().IsEmpty(a1))
                            {
                                world.IssueOrder(new Order("Move", a1, false) { TargetLocation = enemy.CenterLocation.ToCPos() });
                                world.IssueOrder(new Order("Unload", a1, true) { TargetLocation = enemy.CenterLocation.ToCPos() });
                            }
                            else
                                world.IssueOrder(new Order("ReturnToBase", a1, false));
                        }

                        else if (a1.Info.Name.Equals("hind") || a1.Info.Name.Equals("heli") || a1.Info.Name.Equals("yak") || a1.Info.Name.Equals("mig")) // AIR: MOVE TO / ATTACK / RETURN TO BASE (VENI VIDI VICI)
                        {
                            if (a1.TraitOrDefault<LimitedAmmo>().HasAmmo())
                            {
                                var CenterLocation = new CPos(enemy.CenterLocation.ToCPos().X - 4, enemy.CenterLocation.ToCPos().Y - 4);
                                world.IssueOrder(new Order("Move", a1, false) { TargetLocation = CenterLocation });
                                world.IssueOrder(new Order("Attack", a1, true) { TargetActor = enemy });
                            }
                            else
                                world.IssueOrder(new Order("ReturnToBase", a1, false));
                        }

                        else
                        {
                            world.IssueOrder(new Order("Move", a1, false) { TargetLocation = enemy.CenterLocation.ToCPos() });
                            world.IssueOrder(new Order("Attack", a1, true) { TargetActor = enemy }); // Attack ? AttackMove ?
                        }
                    }

                    if (!foundEnemy) /* I would say that is SAD ! */
                    {
                        /* attackTarget = ChooseEnemyTarget();
                        if (attackTarget != null)
                            foreach (var a in attackForce)
                                if (!a.HasTrait<Helicopter>() || !a.HasTrait<Aircraft>())
                                    TryToMove(a, attackTarget.Value, true);
                         */
                    }
                }
            }
        }

        bool IsRallyPointValid(CPos x)
        {
            // this is actually WRONG as soon as HackyAI is building units with a variety of
            // movement capabilities. (has always been wrong)
            return world.IsCellBuildable(x, rallypointTestBuilding);
        }

        void SetRallyPointsForNewProductionBuildings(Actor self)
        {
            var buildings = self.World.ActorsWithTrait<RallyPoint>()
                .Where(rp => rp.Actor.Owner == p &&
                    !IsRallyPointValid(rp.Trait.rallyPoint)).ToArray();

            foreach (var a in buildings)
            {
                CPos newRallyPoint = ChooseRallyLocationNear(a.Actor.Location);
                world.IssueOrder(new Order("SetRallyPoint", a.Actor, false) { TargetLocation = newRallyPoint });
            }
        }

        //won't work for shipyards...
        CPos ChooseRallyLocationNear(CPos startPos)
        {
            var possibleRallyPoints = world.FindTilesInCircle(startPos, 8).Where(IsRallyPointValid).ToArray();
            if (possibleRallyPoints.Length == 0)
            {
                return startPos;
            }

            return possibleRallyPoints.Random(random);
        }

        CPos? ChooseDestinationNear(Actor a, CPos desiredMoveTarget)
        {
            var move = a.TraitOrDefault<IMove>();
            if (move == null) return null;

            CPos xy;
            int loopCount = 0; //avoid infinite loops.
            int range = 2;
            do
            {
                //loop until we find a valid move location
                xy = new CPos(desiredMoveTarget.X + random.Next(-range, range), desiredMoveTarget.Y + random.Next(-range, range));
                loopCount++;
                range = Math.Max(range, loopCount / 2);
                if (loopCount > 10) return null;
            } while (!move.CanEnterCell(xy) && xy != a.Location);

            return xy;
        }

        //try very hard to find a valid move destination near the target.
        //(Don't accept a move onto the subject's current position. maybe this is already not allowed? )
        bool TryToMove(Actor a, CPos desiredMoveTarget, bool attackMove)
        {
            var xy = ChooseDestinationNear(a, desiredMoveTarget);
            if (xy == null)
                return false;
            world.IssueOrder(new Order(attackMove ? "AttackMove" : "Move", a, false) { TargetLocation = xy.Value });
            return true;
        }

        void DeployMcv(Actor self)
        {
            /* find our mcv and deploy it */
            var mcv = self.World.Actors
                .FirstOrDefault(a => a.Owner == p && a.HasTrait<BaseBuilding>());

            if (mcv != null)
            {
                baseCenter = mcv.Location;
                //Don't transform the mcv if it is a fact
                if (mcv.HasTrait<Mobile>())
                    world.IssueOrder(new Order("DeployTransform", mcv, false));
            }
        }

        internal IEnumerable<ProductionQueue> FindQueues(string category)
        {
            return world.ActorsWithTrait<ProductionQueue>()
                .Where(a => a.Actor.Owner == p && a.Trait.Info.Type == category)
                .Select(a => a.Trait);
        }

        //Build a random unit of the given type. Not going to be needed once there is actual AI...
        void BuildRandom(string category)
        {
            // Pick a free queue
            var queue = FindQueues(category).FirstOrDefault(q => q.CurrentItem() == null);
            if (queue == null)
                return;

            var unit = ChooseRandomUnitToBuild(queue);
            if (unit != null && Info.UnitsToBuild.Any(u => u.Key == unit.Name))
            {
                if (unit.Name == "mig" || unit.Name == "yak")
                {
                    int count = countBuilding("afld", p);
                    var myUnits_plane = p.World.ActorsWithTrait<Plane>().Where(a => a.Actor.Owner == p).Count();
                    if (myUnits_plane >= count)
                    {
                        return;
                    }
                }
                else if (unit.Name == "hind" || unit.Name == "heli" || unit.Name == "tran")
                {
                    int count = countBuilding("hpad", p);
                    var myUnits_copter = p.World.ActorsWithTrait<Helicopter>().Where(a => a.Actor.Owner == p).Count();
                    if (myUnits_copter >= count)
                    {
                        return;
                    }
                }
                else if (unit.Name == "harv")
                {
                    var myUnits_harv = p.World.ActorsWithTrait<Harvester>().Where(a => a.Actor.Owner == p).Count();
                    int count = countBuilding("proc", p);
                    if (myUnits_harv >= count * 2)
                    {
                        return;
                    }
                }
                world.IssueOrder(Order.StartProduction(queue.self, unit.Name, 1));
            }
        }

        public void Damaged(Actor self, AttackInfo e) /* p.Stances[e.Attacker.Owner] == Stance.Enemy | Should avoid rocket man to attack himself lol */
        {
            if (!enabled) return;

            var allyUnits = world.FindUnitsInCircle(e.Attacker.CenterLocation, Game.CellSize * 6).Where(a => a.Owner == p && !a.HasTrait<BaseBuilding>()).ToList();
            var enemyUnits = world.FindUnitsInCircle(e.Attacker.CenterLocation, Game.CellSize * 6).Where(a => a.Owner == e.Attacker.Owner && !e.Attacker.HasTrait<BaseBuilding>()).ToList();
            var allyUnitsAway = self.World.ActorsWithTrait<IMove>().Where(a => a.Actor.Owner == p && !a.Actor.HasTrait<BaseBuilding>()).Select(a => a.Actor).ToList();

            if (Info.ShouldRepairBuildings && self.HasTrait<RepairableBuilding>())
                if (e.DamageState > DamageState.Light && e.PreviousDamageState <= DamageState.Light && p.Stances[e.Attacker.Owner] == Stance.Enemy)
                {
                    if (enemyUnits.Count() > allyUnits.Count() * OverrunRation) // TOTAL OVERRUN
                        foreach (var a in allyUnitsAway)
                        {
                            if (a.Info.Name.Equals("hind") || a.Info.Name.Equals("heli") || a.Info.Name.Equals("yak") || a.Info.Name.Equals("mig") && !a.TraitOrDefault<LimitedAmmo>().HasAmmo())
                                return;
                            if (a.Info.Name.Equals("tran") || a.Info.Name.Equals("harv"))
                                return;
                            world.IssueOrder(new Order("Move", a, true) { TargetLocation = e.Attacker.CenterLocation.ToCPos() });
                            world.IssueOrder(new Order("Attack", a, true) { TargetActor = e.Attacker });
                        }

                    world.IssueOrder(new Order("RepairBuilding", self.Owner.PlayerActor, false) { TargetActor = self });
                }

            if (e.Attacker != null && e.Damage > 0 && p.Stances[e.Attacker.Owner] == Stance.Enemy)
            {
                foreach (var a in allyUnits)
                {
                    if (a.Info.Name.Equals("hind") || a.Info.Name.Equals("heli") || a.Info.Name.Equals("yak") || a.Info.Name.Equals("mig") && !a.TraitOrDefault<LimitedAmmo>().HasAmmo())
                        return;
                    if (a.Info.Name.Equals("tran") || a.Info.Name.Equals("harv"))
                        return;
                    world.IssueOrder(new Order("Attack", a, false) { TargetActor = e.Attacker });
                }
                aggro[e.Attacker.Owner].Aggro += e.Damage;
            }
        }
    }
}